/******************************************************************************
 * MODULE       : sensors.c
 * FUNCTION     : Hardware Abstraction Layer of Sensor service in Android Framework
 * AUTHOR       : Masafumi Seike <masafumi.seike@dsn.rohm.co.jp >
 * PROGRAMMED   : Sensor Application Development Group, ROHM CO.,LTD.
 * MODIFICATION : Modified by ROHM, JUN/08/2012
 * REMARKS      :
 * COPYRIGHT    : Copyright (C) 2008 The Android Open Source Project
 *              : Copyright (C) 2011 - 2012 ROHM CO.,LTD.
 *              : Licensed under the Apache License, Version 2.0 (the "License");
 *              : you may not use this file except in compliance with the License.
 *              : You may obtain a copy of the License at
 *              :
 *              :      http://www.apache.org/licenses/LICENSE-2.0
 *              :
 *              : Unless required by applicable law or agreed to in writing, software
 *              : distributed under the License is distributed on an "AS IS" BASIS,
 *              : WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *              : See the License for the specific language governing permissions and
 *              : limitations under the License.
 *****************************************************************************/

/*****************************************************************************/
/* Constant definition (System names)                                        */
/*****************************************************************************/
#define SENSORS_SERVICE_NAME ("sensors")
#define ROHM_DEV_BH1730      ("/dev/bh1730_iodev")

/*****************************************************************************/
/* Include                                                                   */
/*****************************************************************************/
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <cutils/log.h>
#include <cutils/native_handle.h>
#include <cutils/sockets.h>
#include <hardware/sensors.h>
#include <dirent.h>
#include <linux/input.h>

#include "rohm_bh1730_i2c_if.h"

/*****************************************************************************/
/* Constant definition                                                       */
/*****************************************************************************/

/* for logging */
#define D(...)             LOGD(__VA_ARGS__)
#define E(...)             LOGE(__VA_ARGS__)

/* Sensor IDs and names */
#define MAX_NUM_SENSORS      (1)
#define SUPPORTED_SENSORS    ((1 << MAX_NUM_SENSORS) - 1)
#define ID_BASE              (SENSORS_HANDLE_BASE)
#define ID_LIGHT             (ID_BASE + 0)
#define SENSORS_LIGHT        (1 << ID_LIGHT)
#define ID_CHECK(x)          ((unsigned)((x) - ID_BASE) < MAX_NUM_SENSORS)
#define SENSORS_LIST         SENSOR_(LIGHT, "Ambient-Light-Sensor")
#define BH1730_HAL_VER       (0x010004)
#define MAJOR_VERSION(ver)   ((ver >> 16) & 0xFF)
#define MINOR_VERSION(ver)   ((ver >>  8) & 0xFF)
#define BUILD_VERSION(ver)   ((ver >>  0) & 0xFF)

/*****************************************************************************/
/* Structure declaration                                                     */
/*****************************************************************************/

/* Sensor IDs */
static const struct {
    const char *name;
    int        id;
} _sensorIds[MAX_NUM_SENSORS] = {
#define SENSOR_(x, y) { y, ID_ ## x },
    SENSORS_LIST
#undef  SENSOR_
};

/* Sensor control */
typedef struct SensorControl {
    struct sensors_control_device_t device;
    int                             fd;
    uint32_t                        active_sensors;
    int                             event_delay;
    int                             driver_fp;
} SensorControl;

/* Sensor data */
typedef struct SensorData {
    struct sensors_data_device_t device;
    sensors_data_t               sensors[MAX_NUM_SENSORS];
    int                          events_fd;
    uint32_t                     pendingSensors;
    int64_t                      timeStart;
    int64_t                      timeOffset;
} SensorData;

/*****************************************************************************/
/* Prototype declaration for private functions                               */
/*****************************************************************************/
static const char      *_sensorIdToName(int id);
static int             _sensorIdFromName(const char *name);
static int             hal_sensor_open_input_device(char *name_set, int mode);
static native_handle_t *control__open_data_source(struct sensors_control_device_t *dev);
static int             control__activate(struct sensors_control_device_t *dev, int handle, int enabled);
static int             control__set_delay(struct sensors_control_device_t *dev, int32_t ms);
static int             control__wake(struct sensors_control_device_t *dev);
static int             control__close(struct hw_device_t *dev);
static int             data__data_open(struct sensors_data_device_t *dev, native_handle_t *handle);
static int             data__data_close(struct sensors_data_device_t *dev);
static int             pick_sensor(SensorData *data, sensors_data_t *values);
static int             data__poll(struct sensors_data_device_t *dev, sensors_data_t *values);
static int             data__close(struct hw_device_t *dev);
static int             sensors__get_sensors_list(struct sensors_module_t *module, struct sensor_t const **list);
static int             open_sensors(const struct hw_module_t *module, const char *name, struct hw_device_t **device);

/*****************************************************************************/
/* Structure declaration for each module                                     */
/*****************************************************************************/

/* Hardware module */
static struct hw_module_methods_t sensors_module_methods = {
    .open = open_sensors
};

/* Sensors module */
const struct sensors_module_t HAL_MODULE_INFO_SYM = {
    .common = {
        .tag           = HARDWARE_MODULE_TAG,
        .version_major = MAJOR_VERSION(BH1730_HAL_VER),
        .version_minor = MINOR_VERSION(BH1730_HAL_VER),
        .id            = SENSORS_HARDWARE_MODULE_ID,
        .name          = "ROHM SENSORS MODULE",
        .author        = "ROHM SEMICONDUCTOR",
        .methods       = &sensors_module_methods,
    },
    .get_sensors_list = sensors__get_sensors_list
};

/* All supported sensors list */

/** MODULE REGISTRATION SUPPORT
 **
 ** This is required so that hardware/libhardware/hardware.c
 ** will dlopen() this library appropriately.
 **/

/*
 * the following is the list of all supported sensors.
 * this table is used to build sSensorList declared below
 * according to which hardware sensors are reported as
 * available from the emulator (see get_sensors_list below)
 *
 * note: numerical values for maxRange/resolution/power were
 *       taken from the reference AK8976A implementation
 */
static const struct sensor_t sSensorListInit[] = {
    { .name       = "BH1730 Ambient Light Sensor",
      .vendor     = "ROHM SEMICONDUCTOR",
      .version    = 1,
      .handle     = ID_LIGHT,
      .type       = SENSOR_TYPE_LIGHT,
      .maxRange   = 100000,
      .resolution = 0.001,
      .power      = 0.2,
      .reserved   = {}},
};

/* Sensor list */
static struct sensor_t sSensorList[MAX_NUM_SENSORS];

/*****************************************************************************/
/* Functions                                                                 */
/*****************************************************************************/

/******************************************************************************
 * NAME       : _sensorIdToName
 * FUNCTION   : get sensor name from sensor ID
 * REMARKS    :
 *****************************************************************************/
static const char *_sensorIdToName(int id)
{
    int nn;

    for (nn = 0; nn < MAX_NUM_SENSORS; nn++) {
        if (id == _sensorIds[nn].id) {
            return (_sensorIds[nn].name);
        }
    }

    return ("<UNKNOWN>");
}

/******************************************************************************
 * NAME       : _sensorIdFromName
 * FUNCTION   : get sensor ID from sensor name
 * REMARKS    :
 *****************************************************************************/
static int _sensorIdFromName(const char *name)
{
    int nn;

    if (name == NULL) {
        return (-1);
    }
    for (nn = 0; nn < MAX_NUM_SENSORS; nn++) {
        if (!strcmp(name, _sensorIds[nn].name)) {
            return (_sensorIds[nn].id);
        }
    }

    return (-1);
}

/******************************************************************************
 * NAME       : hal_sensor_open_input_device
 * FUNCTION   : get event file pointer
 * REMARKS    :
 *****************************************************************************/
static int hal_sensor_open_input_device(char *name_set, int mode)
{
#define NAME_SPACE (80)

    int           fd;
    char          devname[PATH_MAX];
    char          *filename;
    DIR           *dir;
    struct dirent *de;
    char          name[NAME_SPACE];
    const char    *dirname = "/dev/input";

    fd  = -1;
    dir = opendir(dirname);
    if (dir == NULL) {
        return (-1);
    }
    strcpy(devname, dirname);
    filename    = devname + strlen(devname);
    *filename++ = '/';
    while ((de = readdir(dir))) {
        if ((de->d_name[0] == '.') && \
            ((de->d_name[1] == '\0') || \
             ((de->d_name[1] == '.') && (de->d_name[2] == '\0')))) {
            continue;
        }
        strcpy(filename, de->d_name);
        fd = open(devname, mode);
        if (fd >= 0) {
            if (ioctl(fd, EVIOCGNAME(sizeof(name) - 1), &name) < 1) {
                name[0] = '\0';
            }
            if (!strcmp(name, name_set)) {
                break;
            }
            close(fd);
            fd = -1;
        }
    }
    closedir(dir);

    if (fd < 0) {
        E("couldn't find or open 'compass' driver (%s)", strerror(errno));
    }

    return (fd);

#undef NAME_SPACE
}

/******************************************************************************
 * NAME       : control__open_data_source
 * FUNCTION   : make handle data
 * REMARKS    : this must return a file descriptor that will be used to read
 *            : the sensors data (it is passed to data__data_open() below
 *****************************************************************************/
static native_handle_t *control__open_data_source(struct sensors_control_device_t *dev)
{
#define FILE_LENGTH (50)

    SensorControl   *ctl = (void *)dev;
    native_handle_t *handle;
    int             state;
    char            name_set[FILE_LENGTH];

    memset(name_set, (int)0, sizeof(name_set));
    strcpy(name_set, "bh1730_inputdev");
    ctl->fd = hal_sensor_open_input_device(name_set, O_RDONLY);
    if (ctl->fd < 0) {
        return (NULL);
    }

    /* default */
    handle          = native_handle_create(1, 0);
    handle->data[0] = ctl->fd;

    return (handle);

#undef FILE_LENGTH
}

/******************************************************************************
 * NAME       : control__activate
 * FUNCTION   : check active sensor and get ioctl pointer
 * REMARKS    :
 *****************************************************************************/
static int control__activate(struct sensors_control_device_t *dev, int handle, int enabled)
{
    SensorControl   *ctl = (void *)dev;
    uint32_t        mask, sensors, active, new_sensors, changed;
    int             result;
    POWERON_ALS_ARG pwr_als;

    if ((handle < ID_BASE) || (handle >= ID_BASE + MAX_NUM_SENSORS)) {
        E("%s: bad handle ID", __FUNCTION__);
        return (-1);
    }

    mask        = (1 << handle);
    sensors     = enabled ? mask : 0;
    active      = ctl->active_sensors;
    new_sensors = (active & ~mask) | (sensors & mask);
    changed     = active ^ new_sensors;

    if (!changed) {
        return (0);
    }
    ctl->active_sensors = new_sensors;

    /* open device */
    if (ctl->driver_fp < 0) {
        ctl->driver_fp = open(ROHM_DEV_BH1730, O_RDWR);
        if (ctl->driver_fp < 0) {
            E("device open NG !\n");
            E("ctl->driver_fp = 0x%x !\n", ctl->driver_fp);
            result = (int)ctl->driver_fp;
            return (result);
        }
    }
    switch (handle) {
    case ID_LIGHT:
        if (0 == sensors) {
            pwr_als.power_als = ALS_DISABLE;
        } else {
            pwr_als.power_als = ALS_ENABLE;
        }
        pwr_als.intr = ALS_DISABLE;
        result       = ioctl(ctl->driver_fp, IOCTL_APP_SET_PWRSET_ALS, &pwr_als);
        if (result < 0) {
            E("device ioctl NG !\n");
            E("error value = 0x%x !\n", result);
            return (result);
        }
        break;

    default:
        break;
    }

    return (0);
}

/******************************************************************************
 * NAME       : control__set_delay
 * FUNCTION   : set delay timer
 * REMARKS    :
 *****************************************************************************/
static int control__set_delay(struct sensors_control_device_t *dev, int32_t ms)
{
#define TIME_OFFSET  (125)

    SensorControl *ctl = (void *)dev;
    int           result;

    /* set the given value from application */
    ctl->event_delay = ms + TIME_OFFSET;
    result = ioctl(ctl->driver_fp, IOCTL_APP_SET_TIMER, &(ctl->event_delay));
    if (result < 0) {
        E("SET_TIMER of ioctl is issue NG by device !\n");
        E("error value = 0x%x !\n", result);
        return (result);
    }

    return (0);

#undef TIME_OFFSET
}

/******************************************************************************
 * NAME       : control__wake
 * FUNCTION   : wake control system
 * REMARKS    : this function is used to force-stop the blocking read() in
 *            : data__poll. In order to keep the implementation as simple
 *            : as possible here, we send a command to the emulator which
 *            : shall send back an appropriate data block to the system.
 *****************************************************************************/
static int control__wake(struct sensors_control_device_t *dev)
{
    return (-1);
}

/******************************************************************************
 * NAME       : control__close
 * FUNCTION   : close control system
 * REMARKS    :
 *****************************************************************************/
static int control__close(struct hw_device_t *dev)
{
    int result;

    SensorControl *ctl = (void *)dev;
    /* close device */
    result = ioctl(ctl->driver_fp, IOCTL_APP_SET_TIMER, 0);
    if (result < 0) {
        E("device ioctl NG !\n");
        E("error value = 0x%x !\n", result);
    }
    if (ctl->driver_fp > 0) {
        /* set initialization of timer */
        close(ctl->driver_fp);
    }
    close(ctl->fd);
    free((void *)ctl);

    return (0);
}

/******************************************************************************
 * NAME       : data__data_open
 * FUNCTION   : open data system
 * REMARKS    :
 *****************************************************************************/
static int data__data_open(struct sensors_data_device_t *dev, native_handle_t *handle)
{
    SensorData *data = (void *)dev;
    int        i;

    memset(&data->sensors, 0, sizeof(data->sensors));

    for (i = 0; i < MAX_NUM_SENSORS; i++) {
        /*
           by default all sensors have high accuracy
           (we do this because we don't get an update if the value doesn't
           change).
         */
        data->sensors[i].vector.status = SENSOR_STATUS_ACCURACY_HIGH;
    }
    data->pendingSensors = 0;
    data->events_fd      = dup(handle->data[0]);
    native_handle_close(handle);
    native_handle_delete(handle);

    return (0);
}

/******************************************************************************
 * NAME       : data__data_close
 * FUNCTION   : close data file pointer
 * REMARKS    :
 *****************************************************************************/
static int data__data_close(struct sensors_data_device_t *dev)
{
    SensorData *data = (void *)dev;

    if (data->events_fd > 0) {
        close(data->events_fd);
        data->events_fd = -1;
    }

    return (0);
}

/******************************************************************************
 * NAME       : pick_sensor
 * FUNCTION   : check the pending sensors
 * REMARKS    :
 *****************************************************************************/
static int pick_sensor(SensorData *data, sensors_data_t *values)
{
#define PICK_SEN_WAIT (100000)

    uint32_t mask = SUPPORTED_SENSORS;

    while (mask) {
        uint32_t i = 31 - __builtin_clz(mask);
        mask &= ~(1 << i);
        if (data->pendingSensors & (1 << i)) {
            data->pendingSensors &= ~(1 << i);
            *values               = data->sensors[i];
            values->sensor        = (1 << i);
            return (i);
        }
    }
    E("no sensor to return!!! pendingSensors=%08x", data->pendingSensors);
    /* we may end-up in a busy loop, slow things down, just in case. */
    usleep(PICK_SEN_WAIT);

    return (-1);

#undef PICK_SEN_WAIT
}

/******************************************************************************
 * NAME       : data__poll
 * FUNCTION   : handle value of Ambit Light Sensor
 * REMARKS    :
 *****************************************************************************/
static int data__poll(struct sensors_data_device_t *dev, sensors_data_t *values)
{
#define SECOND_TIME (1000000000)
#define MICRO_TIME  (1000)
#define CUT_UNIT    (1000)

    SensorData *data = (void *)dev;

    unsigned char *read_data;
    float         calc_data;
    long          cut_data;
    int           fd = data->events_fd;

    if (fd < 0) {
        E("invalid file descriptor, fd=%d", fd);
        return (-1);
    }

    /* there are pending sensors, returns them now... */
    if (data->pendingSensors) {
        return (pick_sensor(data, values));
    }
    /* wait until we get a complete event for an enabled sensor */
    uint32_t new_sensors = 0;
    while (1) {
        /* read the next event */
        struct input_event event;
        int                nread;

        /* read the sensor and timer value from event */
        nread = read(fd, &event, sizeof(event));
        if (nread == sizeof(event)) {
            uint32_t v;
            if (event.type == EV_ABS) {
                switch (event.code) {
                case ABS_X:
                    new_sensors                  |= SENSORS_LIGHT;
                    calc_data                     = (float)event.value / (1 << DECIMAL_BIT);
                    cut_data                      = (long)(calc_data * CUT_UNIT);
                    data->sensors[ID_LIGHT].light = (float)cut_data / CUT_UNIT;
                    break;
                default:
                    break;
                }
            } else if (event.type == EV_SYN) {
                if (event.code == SYN_CONFIG) {
                    /*
                       we use SYN_CONFIG to signal that we need to exit the main loop.
                       LOGD("got empty message: value=%d", event.value);
                     */
                    return (0x7FFFFFFF);
                }
                if (new_sensors) {
                    data->pendingSensors = new_sensors;
                    int64_t t = event.time.tv_sec * SECOND_TIME + event.time.tv_usec * MICRO_TIME;
                    while (new_sensors) {
                        uint32_t i = 31 - __builtin_clz(new_sensors);
                        new_sensors &= ~(1 << i);
                        data->sensors[i].time = t;
                    }
                    return (pick_sensor(data, values));
                }
            }
        }
    }

#undef SECOND_TIME
#undef MICRO_TIME
#undef CUT_UNIT
}

/******************************************************************************
 * NAME       : data__close
 * FUNCTION   : close data system
 * REMARKS    :
 *****************************************************************************/
static int data__close(struct hw_device_t *dev)
{
    SensorData *data = (SensorData *)dev;

    if (data) {
        if (data->events_fd > 0) {
            close(data->events_fd);
        }
        free(data);
    }

    return (0);
}

/******************************************************************************
 * NAME       : sensors__get_sensors_list
 * FUNCTION   : get all sensors list
 * REMARKS    :
 *****************************************************************************/
static int sensors__get_sensors_list(struct sensors_module_t *module, struct sensor_t const **list)
{
    *list = sSensorListInit;
    return (sizeof(sSensorList) / sizeof(sSensorList[0]));
}

/******************************************************************************
 * NAME       : open_sensors
 * FUNCTION   : initialize system
 * REMARKS    :
 *****************************************************************************/
static int open_sensors(const struct hw_module_t *module, const char *name, struct hw_device_t **device)
{
    int status = -EINVAL;

    if (!strcmp(name, SENSORS_HARDWARE_CONTROL)) {
        SensorControl *dev = malloc(sizeof(*dev));

        memset(dev, 0, sizeof(*dev));

        dev->device.common.tag       = HARDWARE_DEVICE_TAG;
        dev->device.common.version   = 0;
        dev->device.common.module    = (struct hw_module_t *) module;
        dev->device.common.close     = control__close;
        dev->device.open_data_source = control__open_data_source;
        dev->device.activate         = control__activate;
        dev->device.set_delay        = control__set_delay;
        dev->device.wake             = control__wake;
        dev->fd                      = -1;
        dev->driver_fp               = -2;

        *device = &dev->device.common;
        status  = 0;
    } else if (!strcmp(name, SENSORS_HARDWARE_DATA)) {
        SensorData *dev = malloc(sizeof(*dev));

        memset(dev, 0, sizeof(*dev));

        dev->device.common.tag     = HARDWARE_DEVICE_TAG;
        dev->device.common.version = 0;
        dev->device.common.module  = (struct hw_module_t *) module;
        dev->device.common.close   = data__close;
        dev->device.data_open      = data__data_open;
        dev->device.data_close     = data__data_close;
        dev->device.poll           = data__poll;
        dev->events_fd             = -1;

        *device = &dev->device.common;
        status  = 0;
    }
    return (status);
}

